#include <glib/gi18n.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <stdlib.h>
#include <unistd.h>
#include "gyum_utils.h"
#include "yumrepo.h"

#ifndef YUM_CONFIG
#define YUM_CONFIG		"/etc/yum.conf"
#endif

static gchar *yumrepo_cache_root=NULL;
static gchar *yumrepo_log_file=NULL;
static gchar *yumrepo_yum_repos_d=NULL;

RepoStruct *yumrepo_struct_alloc()
{
	RepoStruct *rs=g_new0(RepoStruct,1);
	rs->change=FALSE;
	rs->remove=FALSE;
	rs->vars=g_hash_table_new_full(g_str_hash,g_str_equal,g_free,g_free);
	return rs;
}

void yumrepo_struct_copy_vars(const gchar *var,const gchar *val,GHashTable *ht)
{
	g_hash_table_insert(ht,g_strdup(var),g_strdup(val));
}

void yumrepo_struct_copy(RepoStruct *src,RepoStruct *dest)
{
	dest->remove=src->remove;
	dest->change=src->change;
	dest->filename=g_strdup(src->filename);
	g_hash_table_foreach(src->vars,(GHFunc)yumrepo_struct_copy_vars,dest->vars);
}

void yumrepo_struct_free(RepoStruct *rs)
{
	g_hash_table_destroy(rs->vars);
	if (rs->filename)
		g_free(rs->filename);
	g_free(rs);
}

void yumrepo_struct_set_id(RepoStruct *rs,const gchar *id)
{
	if (access(rs->filename,F_OK))
	{
		if (rs->filename) g_free(rs->filename);
		rs->filename=g_strdup_printf("%s/%s.repo",yumrepo_get_yum_repos_d(),id);
		g_warning("rs->filename: %s",rs->filename);
	}
}

static gboolean yumrepo_loading;

void yumrepo_struct_set_variable(RepoStruct *rs,const gchar *var,const gchar *val)
{
	gchar *old_value=NULL;
	if (!yumrepo_loading) old_value=g_hash_table_lookup(rs->vars,var);
	if (!old_value || strcmp(old_value,val))
	{
		if (!yumrepo_loading) rs->change=TRUE;
		g_hash_table_replace(rs->vars,g_strdup(var),g_strdup(val));
	}
}

gchar *yumrepo_filename(const gchar *id)
{
	gint i=0;
	gchar *fullname;
	fullname=g_strdup_printf("%s/%s.repo",yumrepo_get_yum_repos_d(),id);
	while (!access(fullname,F_OK))
	{
		g_free(fullname);
		i++;
		fullname=g_strdup_printf("%s/%s%i.repo",yumrepo_get_yum_repos_d(),id,i);
	}
	return fullname;
}

void yumrepo_parse_line(GHashTable *yumrepo_list,gchar **id,RepoStruct **repo,gchar *buffer,gchar *fullname)
{
	static RepoStruct *yumrepo_prev_repo=NULL;
	static gchar *yumrepo_prev_var=NULL;
	gchar *ptr;
	gchar *var;
	if (yumrepo_prev_repo!=*repo)
	{
		yumrepo_prev_repo=*repo;
		if (yumrepo_prev_var) g_free(yumrepo_prev_var);
		yumrepo_prev_var=NULL;
	}
	ptr=strchr(buffer,'\n');
	if (ptr) *ptr=0;
	ptr=buffer;
	while ((*ptr==' ') || (*ptr=='\t')) ptr++;
	var=strchr(ptr,'=');
	if (var)
	{
		gchar *val=var;
		var--;
		while ((*var==' ') || (*var=='\t')) strcpy(var,var+1);
		while ((*val==' ') || (*val=='\t')) strcpy(val,val+1);
		var=ptr;
		val=strchr(ptr,'=');
		*val=0;
		val++;
		if (*repo)
		{
			if (yumrepo_prev_var) g_free(yumrepo_prev_var);
			yumrepo_struct_set_variable(*repo,var,val);
			yumrepo_prev_var=g_strdup(var);
		}
	}
	else
	if (*ptr=='[')
	{
		var=strchr(ptr,']');
		if (var)
		{
			*var=0;
			var=ptr+1;
			if (*repo)
			{
				g_hash_table_insert(yumrepo_list,*id,*repo);
			}
			*id=g_strdup(var);
			*repo=yumrepo_struct_alloc();
			(*repo)->change=FALSE;
			yumrepo_prev_repo=*repo;
			if (yumrepo_prev_var) g_free(yumrepo_prev_var);
			yumrepo_prev_var=NULL;
			if (!fullname)
			{
				if (!(*repo)->filename)
				{
					fullname=yumrepo_filename(*id);
					(*repo)->filename=fullname;
				}
			}
			else
				(*repo)->filename=g_strdup(fullname);
		}
	}
	else
	{
		if (!*ptr || (*ptr=='#'))
		{
			if (yumrepo_prev_var) g_free(yumrepo_prev_var);
			yumrepo_prev_var=NULL;
		}
		else
		if (yumrepo_prev_var)
		{
			gchar *new_val;
			gchar *old_val;
			old_val=g_hash_table_lookup((*repo)->vars,yumrepo_prev_var);
			new_val=g_strdup_printf("%s %s",old_val,ptr);
			yumrepo_struct_set_variable(*repo,yumrepo_prev_var,new_val);
		}
	}
}

void yumrepo_load_text(GHashTable *yumrepo_list,const gchar *text)
{
	RepoStruct *repo=NULL;
	gchar *id=NULL;
	gchar *ptr=g_strdup(text);
	while (ptr && *ptr)
	{
		gchar *line=ptr;
		ptr=strchr(ptr,'\n');
		if (ptr)
		{
			*ptr=0;
			ptr++;
		}
		yumrepo_parse_line(yumrepo_list,&id,&repo,line,NULL);
		repo->change=TRUE;
	}
	if (repo)
	{
		g_hash_table_insert(yumrepo_list,id,repo);
	}
}

void yumrepo_load_file(GHashTable *yumrepo_list,const gchar *dir,const gchar *filename)
{
	gchar *fullname;
	RepoStruct *repo=NULL;
	gchar *id=NULL;
	FILE *file;
	fullname=g_strdup_printf("%s/%s",dir,filename);
	file=fopen(fullname,"r");
	if (file)
	{
		gchar buffer[1024];
		yumrepo_loading=TRUE;
		while (!feof(file) && !ferror(file) && fgets(buffer,1024,file))
		{
			yumrepo_parse_line(yumrepo_list,&id,&repo,buffer,fullname);
		}
		if (repo)
		{
			g_hash_table_insert(yumrepo_list,id,repo);
		}
		fclose(file);
		yumrepo_loading=FALSE;
	}
	else
		yumrepo_struct_free(repo);
	g_free(fullname);
}

GHashTable *yumrepo_load()
{
	GHashTable *yumrepo_list;
	gchar *dn=g_path_get_dirname(YUM_CONFIG);
	gchar *fn=g_path_get_basename(YUM_CONFIG);
	DIR *dir;
	yumrepo_list=g_hash_table_new_full(g_str_hash,g_str_equal,g_free,(GDestroyNotify)yumrepo_struct_free);
	yumrepo_load_file(yumrepo_list,dn,fn);
	g_free(dn);
	g_free(fn);
	dir=opendir(yumrepo_get_yum_repos_d());
	if (dir)
	{
		struct dirent *de;
		while ((de=readdir(dir))!=NULL)
		{
			gchar *ptr;
			ptr=strstr(de->d_name,".repo");
			if (ptr && !ptr[5])
				yumrepo_load_file(yumrepo_list,yumrepo_get_yum_repos_d(),de->d_name);
		}
		closedir(dir);
	}
	return yumrepo_list;
}

void yumrepo_remove(RepoStruct *repo)
{
	gchar *cmd;
	cmd=g_strdup_printf("rm -f %s",repo->filename);
	gyum_command_execute(_("Removing repository"),cmd,FALSE,NULL,NULL);
	g_free(cmd);
}

static gboolean check_paths=FALSE;

void yumrepo_save_var(const gchar *var,const gchar *val,FILE *file)
{
	fprintf(file,"%s=%s\n",var,val);
	if (check_paths && !strcmp(var,"cachedir"))
	{
		if (!yumrepo_cache_root || strcmp(yumrepo_cache_root,val))
		{
			if (yumrepo_cache_root) g_free(yumrepo_cache_root);
			yumrepo_cache_root=g_strdup(val);
		}
	}
	if (check_paths && !strcmp(var,"logfile"))
	{
		if (!yumrepo_log_file || strcmp(yumrepo_log_file,val))
		{
			if (yumrepo_log_file) g_free(yumrepo_log_file);
			yumrepo_log_file=g_strdup(val);
		}
	}
	if (check_paths && !strcmp(var,"reposdir"))
	{
		if (!yumrepo_yum_repos_d || strcmp(yumrepo_yum_repos_d,val))
		{
			if (yumrepo_yum_repos_d) g_free(yumrepo_yum_repos_d);
			yumrepo_yum_repos_d=g_strdup(val);
		}
	}
}

static gboolean yumrepo_change=FALSE;

void yumrepo_save(const gchar *id,RepoStruct *repo)
{
	FILE *file;
	file=fopen(repo->filename,"a");
	if (file)
	{
		yumrepo_change=TRUE;
		fprintf(file,"[%s]\n",id);
		check_paths=!strcmp(id,"main");
		g_hash_table_foreach(repo->vars,(GHFunc)yumrepo_save_var,file);
		fprintf(file,"\n");
		fclose(file);
	}
}

static GList *yumrepo_file_backup;

void yumrepo_unload_one_backup(const gchar *id,RepoStruct *repo,GList **yumrepo_unload_backup_error)
{
	if (!*yumrepo_unload_backup_error &&
		!access(repo->filename,F_OK) &&
		repo->change &&
		!repo->remove &&
		!g_list_find_custom(yumrepo_file_backup,repo->filename,(GCompareFunc)strcmp))
	{
		gchar *cmd;
		cmd=g_strdup_printf("mv -f %s %s.save",repo->filename,repo->filename);
		if (!gyum_command_execute(_("Saving original repository"),cmd,FALSE,NULL,NULL))
			*yumrepo_unload_backup_error=g_list_append(*yumrepo_unload_backup_error,(gpointer)id);
		else
			yumrepo_file_backup=g_list_append(yumrepo_file_backup,repo->filename);
		g_free(cmd);
	}
}

void yumrepo_unload_one(const gchar *id,RepoStruct *repo,GList *yumrepo_unload_backup_error)
{
	if (repo->remove)
	{
		yumrepo_remove(repo);
		yumrepo_change=TRUE;
	}
	else
	{
		if ((repo->change ||
			g_list_find_custom(yumrepo_file_backup,repo->filename,(GCompareFunc)strcmp)) &&
			!g_list_find_custom(yumrepo_unload_backup_error,id,(GCompareFunc)strcmp))
		{
			yumrepo_save(id,repo);
		}
	}
}

gboolean yumrepo_unload(GHashTable *yumrepo_list,gboolean update_disk,gboolean *changed)
{
	gboolean successful=TRUE;
	yumrepo_change=FALSE;
	if (yumrepo_list)
	{
		if (update_disk)
		{
			GList *yumrepo_unload_backup_error=NULL;
			yumrepo_file_backup=NULL;
			g_hash_table_foreach(yumrepo_list,(GHFunc)yumrepo_unload_one_backup,&yumrepo_unload_backup_error);
			if (!yumrepo_unload_backup_error)
				g_hash_table_foreach(yumrepo_list,(GHFunc)yumrepo_unload_one,yumrepo_unload_backup_error);
			else
				successful=FALSE;
			g_list_free(yumrepo_unload_backup_error);
			g_list_free(yumrepo_file_backup);
		}
	}
	if (changed) *changed|=yumrepo_change;
	return successful;
}

void yumrepo_preload_paths()
{
	if (!yumrepo_cache_root || !yumrepo_log_file)
	{
		GHashTable *yumrepo_list;
		gchar *dn=g_path_get_dirname(YUM_CONFIG);
		gchar *fn=g_path_get_basename(YUM_CONFIG);
		gchar *data;
		RepoStruct *rs;
		yumrepo_list=g_hash_table_new_full(g_str_hash,g_str_equal,g_free,(GDestroyNotify)yumrepo_struct_free);
		yumrepo_load_file(yumrepo_list,dn,fn);
		g_free(dn);
		g_free(fn);
		rs=g_hash_table_lookup(yumrepo_list,"main");
		if (rs)
		{
			data=g_hash_table_lookup(rs->vars,"cachedir");
			if (data) yumrepo_cache_root=g_strdup(data);
			else yumrepo_cache_root=g_strdup("/var/cache/yum");
			data=g_hash_table_lookup(rs->vars,"logfile");
			if (data) yumrepo_log_file=g_strdup(data);
			else yumrepo_log_file=g_strdup("/var/log/yum.log");
			data=g_hash_table_lookup(rs->vars,"reposdir");
			if (data) yumrepo_yum_repos_d=g_strdup(data);
			else yumrepo_yum_repos_d=g_strdup("/etc/yum.repos.d");
		}
		g_hash_table_destroy(yumrepo_list);
	}
}

const gchar *yumrepo_get_cache_root()
{
	yumrepo_preload_paths();
	return yumrepo_cache_root;
}

const gchar *yumrepo_get_log_file()
{
	yumrepo_preload_paths();
	return yumrepo_log_file;
}

const gchar *yumrepo_get_yum_repos_d()
{
	yumrepo_preload_paths();
	return yumrepo_yum_repos_d;
}
